from pathlib import Path
from shutil import which

import pytest

from diopter.compiler import (
    CompilerExe,
    CompilerProject,
    Language,
    SourceProgram,
    parse_compiler,
)
from diopter.sanitizer import Sanitizer


def find_clang() -> CompilerExe | None:
    v = ["-10", "-11", "-12", "-13", "-14", "-15", "-16", ""]
    v.reverse()

    for version in v:
        path = Path("clang" + version)
        project_revision = parse_compiler(path)
        assert project_revision
        if which(path):
            return CompilerExe(CompilerProject.LLVM, path, project_revision[1])

    return None


def test_check_for_compiler_warnings() -> None:
    # TODO: Can I find a test case that only clang
    # catches and a test case that only gcc catches?

    clang = find_clang()
    assert clang, "Could not find a clang executable"
    san = Sanitizer(clang=clang)
    p1 = SourceProgram(
        code="int main(){return 0;}",
        language=Language.C,
    )

    assert san.check_for_compiler_warnings(p1)

    p2 = SourceProgram(
        code="void main(){}",
        language=Language.C,
    )

    assert san.check_for_compiler_warnings(p2).check_warnings_failed

    p2 = SourceProgram(
        code="v main(){}",
        language=Language.C,
    )
    assert san.check_for_compiler_warnings(p2).check_warnings_failed


@pytest.mark.parametrize(
    "code",
    [
        "int main(){ int a[1] = {0}; return a[1];}",
        "int printf(const char *, ...); int main()"
        '{int a = 2147483647;printf("a+1: %d", a+1);}',
        "int *g;" "void foo(int x) { g = &x;}" "int main(){foo(0); return *g;}",
    ],
)
def test_ub_and_address_sanitizer(code: str) -> None:
    clang = find_clang()
    assert clang, "Could not find a clang executable"
    san = Sanitizer(clang=clang, debug=True)
    p = SourceProgram(code=code, language=Language.C)
    assert san.check_for_sanitizer_errors(
        p, sanitizer_flag="undefined,address"
    ).sanitizer_failed


@pytest.mark.parametrize(
    "code",
    [
        "int main(){ int a[1]; if (a[0]) return 1; return 0;}",
    ],
)
def test_memory_sanitizer(code: str) -> None:
    clang = find_clang()
    assert clang, "Could not find a clang executable"
    san = Sanitizer(
        clang=clang,
        use_ub_address_sanitizer=False,
        use_memory_sanitizer=True,
        debug=True,
    )
    p = SourceProgram(code=code, language=Language.C)
    assert san.check_for_sanitizer_errors(p, sanitizer_flag="memory").sanitizer_failed


if __name__ == "__main__":
    test_check_for_compiler_warnings()



from pathlib import Path
from tempfile import TemporaryDirectory

from diopter.bisector import BisectionCallback, bisect, rev_parse
from diopter.repository import Commit, Repo, Revision
from diopter.utils import run_cmd


class TestBisectionCallback(BisectionCallback):
    def check_impl(self, commit: Commit, repo_dir: Path) -> bool | None:
        filenames = list(f for f in repo_dir.iterdir() if f.name != ".git")
        if len(filenames) == 4:
            # Trigger a bisect skip
            return None
        return len(filenames) > 2


def test_bisection() -> None:
    with TemporaryDirectory() as tmpdir:
        run_cmd(f"git -C {tmpdir} init")
        commits = []
        for f in ("a", "b", "c", "d", "e", "f"):
            fpath = Path(tmpdir) / f
            fpath.touch()
            run_cmd(f"git -C {tmpdir} add {f}")
            run_cmd(f"git -C {tmpdir} commit -m 'Add {f}'")
            commits.append(rev_parse(Path(tmpdir), "HEAD"))

        repo = Repo(Path(tmpdir), Revision("master"))
        bad = commits[-1]
        good = commits[0]
        callback = TestBisectionCallback()
        result = bisect(
            repo,
            good=good,
            bad=bad,
            callback=callback,
            no_checkout=False,
        )
        assert result is not None
        assert result == commits[2]



from __future__ import annotations

import subprocess
from abc import ABC, abstractmethod
from concurrent.futures import Executor
from itertools import repeat
from pathlib import Path
from random import randint
from shutil import which
from typing import Iterator

from diopter.compiler import Language, SourceProgram
from diopter.sanitizer import Sanitizer
from diopter.utils import TempDirEnv


def dummy_func(generator: Generator) -> SourceProgram:
    return generator.generate_program()


class Generator(ABC):
    def __init__(self, sanitizer: Sanitizer):
        self.sanitizer = sanitizer

    @abstractmethod
    def generate_program_impl(self) -> SourceProgram:
        """Concrete subclasses must implement this

        Returns:
            SourceProgram: generated program
        """
        pass

    @abstractmethod
    def filter_program(self, program: SourceProgram) -> bool:
        """Concrete subclasses must implement this to check if the generated
           program should be discared
        Args:
            program (SourceProgram): the program to check
        Returns:
            bool: if the program is good(sanitized)
        """
        pass

    def generate_program(self) -> SourceProgram:
        while True:
            program = self.generate_program_impl()
            if self.filter_program(program):
                return program

    def generate_programs_parallel(
        self, n: int, executor: Executor, chunksize: int = 10
    ) -> Iterator[SourceProgram]:
        """
        Generate programs in parallel. Yield futures wrapping the generation
        jobs.

        Example:
        with ProcessPoolExecutor(16) as executor:
            for fut in generator.generate_programs_parallel(100, executor):
                program = fut.result()

        Args:
            n (int):
                how many cases to generate
            executor (Executor):
                executor used for running the code generation jobs
        Returns:
            Iterator[SourceProgram]: the generated programs
        """
        return executor.map(dummy_func, repeat(self, n), chunksize=chunksize)


def find_csmith_include_path() -> str:
    """Find csmith include path.

    Returns:
        str: csmith include path
    """
    if Path("/usr/include/csmith-2.3.0").exists():
        return "/usr/include/csmith-2.3.0"

    if Path("/usr/include/csmith").exists():
        return "/usr/include/csmith"

    if Path("/usr/include/csmith.h").exists():
        return "/usr/include/"

    raise RuntimeError("Could not find csmith include path")


class CSmithGenerator(Generator):
    default_options_pool = [
        "arrays",
        "bitfields",
        "comma-operators",
        "compound-assignment",
        "consts",
        "divs",
        "embedded-assigns",
        "jumps",
        "longlong",
        "force-non-uniform-arrays",
        "math64",
        "muls",
        "packed-struct",
        "paranoid",
        "pointers",
        "structs",
        "inline-function",
        "return-structs",
        "arg-structs",
        "dangling-global-pointers",
    ]
    fixed_options = [
        "--no-unions",
        "--safe-math",
        "--no-argc",
        "--no-volatiles",
        "--no-volatile-pointers",
    ]

    def __init__(
        self,
        sanitizer: Sanitizer,
        csmith: str | None = None,
        include_path: str | None = None,
        options_pool: list[str] | None = None,
        minimum_length: int = 10000,
        maximum_length: int = 50000,
    ):
        """
        Args:
            sanitizer (Sanitizer):
                used to sanitize and discard generated code
            csmith (str | None):
                Path to csmith executable, if empty "csmith" will be used
            include_path (str | None):
                csmith include path, if empty "/usr/include/csmith-2.3.0" will be used
            options_pool (list[str] | None):
                csmith options that will be randomly selected,
                if empty default_options_pool will be used
            minimum_length (int):
                The minimum length of a generated test case in characters.
            maximum_length (int):
                The maximum length of a generated test case in characters.
        """
        super().__init__(sanitizer)
        self.minimum_length = minimum_length
        self.maximum_length = maximum_length
        self.csmith = csmith if csmith else "csmith"
        self.options = (
            options_pool if options_pool else CSmithGenerator.default_options_pool
        )
        self.include_path = include_path if include_path else find_csmith_include_path()

        if not Path(self.include_path).exists():
            raise ValueError(f"Invalid csmith include path: {self.include_path}")

        if not which(self.csmith):
            raise ValueError(f"Invalid csmith executable: {self.csmith}")

    def generate_program_impl(self) -> SourceProgram:
        """Generate random code with csmith.

        Returns:
            SourceProgram: csmith generated program.
        """

        cmd = [self.csmith] + CSmithGenerator.fixed_options
        for option in self.options:
            if randint(0, 1):
                cmd.append(f"--{option}")
            else:
                cmd.append(f"--no-{option}")
        result = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        assert result.returncode == 0
        return SourceProgram(
            code=result.stdout.decode("utf-8"),
            language=Language.C,
            defined_macros=(),
            include_paths=(),
            system_include_paths=(self.include_path,),
            flags=(),
        )

    def filter_program(self, program: SourceProgram) -> bool:
        if (
            len(program.code) < self.minimum_length
            or len(program.code) > self.maximum_length
        ):
            return False
        return bool(self.sanitizer.sanitize(program))


class YarpGen(Generator):
    def __init__(
        self,
        sanitizer: Sanitizer,
        yarpgen: str | None = None,
        language: Language = Language.C,
        additional_flags: tuple[str, ...] = (),
        minimum_length: int = 10000,
        maximum_length: int = 50000,
    ):
        """
        Args:
            sanitizer (Sanitizer):
                used to sanitize and discard generated code
            yarpgen (str | None):
                Path to yarpgen executable, if empty "csmith" will be used
            language (Language):
                The language of the generated code (--std flag).
            additional_flags (tuple[str, ...]):
                Additional flags to pass to yarpgen.
            minimum_length (int):
                The minimum length of a generated test case in characters.
            maximum_length (int):
                The maximum length of a generated test case in characters.
        """
        super().__init__(sanitizer)
        self.minimum_length = minimum_length
        self.maximum_length = maximum_length
        self.yarpgen = yarpgen if yarpgen else "yarpgen"
        self.language = language
        self.additional_flags = additional_flags

        if not which(self.yarpgen):
            raise ValueError(f"Invalid yarpgen executable: {self.yarpgen}")

    def generate_program_impl(self) -> SourceProgram:
        """Generate random code with yarpgen.

        Returns:
            SourceProgram: yarpgen generated program.
        """

        lang_flag = {
            Language.C: "--std=c",
            Language.CPP: "--std=c++",
        }[self.language]
        with TempDirEnv() as temp_dir:
            cmd = [self.yarpgen, lang_flag, "-o", str(temp_dir.resolve())] + list(
                self.additional_flags
            )
            for _ in range(100):
                result = subprocess.run(
                    cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT
                )
                if result.returncode == 0:
                    break
            assert result.returncode == 0, result.stdout.decode("utf-8")
            driver_str = (temp_dir / ("driver" + self.language.to_suffix())).read_text()
            func_str = (
                (temp_dir / ("func" + self.language.to_suffix()))
                .read_text()
                .replace('#include "init.h"', "")
            )

            return SourceProgram(
                code=driver_str + "\n" + func_str,
                language=self.language,
                defined_macros=(),
                include_paths=(),
                system_include_paths=(),
                flags=(),
            )

    def filter_program(self, program: SourceProgram) -> bool:
        if (
            len(program.code) < self.minimum_length
            or len(program.code) > self.maximum_length
        ):
            return False
        return bool(self.sanitizer.sanitize(program))



from __future__ import annotations

import os
import subprocess
from functools import cache
from pathlib import Path
from typing import NewType

from diopter.compiler import CompilerProject
from diopter.utils import run_cmd

DEFAULT_REPOS_DIR = Path.home() / ".cache" / "diopter-compiler-repos"


class RepositoryException(Exception):
    pass


Revision = NewType("Revision", str)
Commit = NewType("Commit", str)


class Repo:
    def __init__(self, path: Path, main_branch: Revision):
        self.path = os.path.abspath(path)
        try:
            run_cmd(f"git -C {self.path} status")
        except Exception:
            raise ValueError(f"{path} is not a git repository")
        self.main_branch = main_branch

    def current_branch(self) -> Revision:
        return Revision(run_cmd(f"git -C {self.path} branch --show-current").stdout)

    @cache
    def get_best_common_ancestor(self, rev_a: Revision, rev_b: Revision) -> Commit:
        a = self.rev_to_commit(rev_a)
        b = self.rev_to_commit(rev_b)
        return Commit(run_cmd(f"git -C {self.path} merge-base {a} {b}").stdout)

    @cache
    def rev_to_commit(self, rev: Revision) -> Commit:
        """Convert any revision (commits, tags etc.) into their
        SHA1 hash via git rev-parse.

        Args:
            rev (str): Revision to convert.

        Returns:
            str: Hash of `rev` in this repo.
        """
        # Could support list of revs...
        try:
            if rev == "trunk" or rev == "master" or rev == "main":
                rev = self.main_branch
            return Commit(run_cmd(f"git -C {self.path} rev-parse {rev}").stdout)
        except subprocess.CalledProcessError as e:
            raise RepositoryException(e)

    def rev_to_range_needing_patch(
        self, introducer: Commit, fixer: Commit
    ) -> list[Commit]:  # noqa: W605
        """
        This function's aim is best described with a picture # noqa: W605
           O---------P
          /   G---H   \      I---J       L--M
         /   /     \   \    /     \     /
        A---B---Z---C---N---D-------E---F---K
             \     /
              Q---R
        call rev_to_range_needing_patch(G, K) gives
        (K, F, 'I, J, D, E', C, H, G)
        in particular it doesn't include Z, P, O, Q and R
        Range G~..K would include these

        Args:
            introducer (Commit): introducer commit
            fixer (Commit): fixer commit

        Returns:
            list[Commit]: List of revision hashes needing the patch.
        """
        #

        # Get all commits with at least 2 parents
        try:
            merges_after_introducer = run_cmd(
                f"git -C {self.path} rev-list --merges {introducer}~..{fixer}",
            ).stdout
        except subprocess.CalledProcessError as e:
            raise RepositoryException(e)

        if len(merges_after_introducer) > 0:
            # Get all parent commits of these (so for C it would be H, Z and R)
            cmd = f"git -C {self.path} rev-parse " + "^@ ".join(merges_after_introducer)
            try:
                merger_parents = set(run_cmd(cmd).stdout.split("\n"))
            except subprocess.CalledProcessError as e:
                raise RepositoryException(e)

            # Remove all parents which are child of the requested commit
            unwanted_merger_parents = [
                parent
                for parent in merger_parents
                if not self.is_ancestor(Revision(introducer), Revision(parent))
            ]
        else:
            unwanted_merger_parents = []
        cmd = f"git -C {self.path} rev-list {fixer} ^{introducer} " + " ^".join(
            unwanted_merger_parents
        )
        try:
            res = [
                Commit(commit)
                for commit in run_cmd(cmd).stdout.split("\n")
                if commit != ""
            ] + [introducer]
            return res
        except subprocess.CalledProcessError as e:
            raise RepositoryException(e)

    def direct_first_parent_path(self, older: Commit, younger: Commit) -> list[Commit]:
        """Get interval of commits [younger, older] always following the
        first parent.

        Args:
            self:
            older (Commit): Older commit
            younger (Commit): Younger commit

        Returns:
            list[Commit]: Commits [younger, older] following the first parent.
        """
        cmd = f"git -C {self.path} rev-list --first-parent {younger} ^{older}"
        try:
            res = [
                Commit(commit)
                for commit in run_cmd(cmd).stdout.split("\n")
                if commit != ""
            ] + [older]
            return res
        except subprocess.CalledProcessError as e:
            raise RepositoryException(e)

    def rev_to_commit_list(self, rev: Revision) -> list[Commit]:
        try:
            return [
                Commit(commit)
                for commit in run_cmd(
                    f"git -C {self.path} log --format=%H {rev}"
                ).stdout.split("\n")
            ]

        except subprocess.CalledProcessError as e:
            raise RepositoryException(e)

    def is_ancestor(
        self, rev_old: Revision | Commit, rev_young: Revision | Commit
    ) -> bool:
        commit_old = self.rev_to_commit(rev_old)
        commit_young = self.rev_to_commit(rev_young)

        process = subprocess.run(
            f"git -C {self.path} merge-base --is-ancestor "
            f"{commit_old} {commit_young}".split(" "),
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        return process.returncode == 0

    def is_branch_point_ancestor_wrt_master(
        self, rev_old: Revision, rev_young: Revision
    ) -> bool:
        """
        In the following example, Young is not ancestor of Old but
        their respective best common ancestors wrt to main (i.e. the commit
        where they branched away from Main) are ancestors.

        Main
         | Young
         |/
         | Old
         |/

        Args:
            self:
            rev_old (Revision): rev_old
            rev_young (Revision): rev_young

        Returns:
            bool: True if their best common ancestors with main are ancestors.
        """
        commit_old = self.rev_to_commit(rev_old)
        commit_young = self.rev_to_commit(rev_young)
        commit_master = self.rev_to_commit("master")
        ca_young = self.get_best_common_ancestor(commit_master, commit_young)
        ca_old = self.get_best_common_ancestor(commit_master, commit_old)

        return self.is_ancestor(ca_old, ca_young)

    def on_same_branch_wrt_master(self, rev_a: Revision, rev_b: Revision) -> bool:
        commit_a = self.rev_to_commit(rev_a)
        commit_b = self.rev_to_commit(rev_b)
        commit_master = self.rev_to_commit("master")

        ca_a = self.get_best_common_ancestor(commit_a, commit_master)
        ca_b = self.get_best_common_ancestor(commit_b, commit_master)

        return ca_b == ca_a

    def get_unix_timestamp(self, rev: Revision) -> int:
        commit = self.rev_to_commit(rev)
        try:
            return int(
                run_cmd(
                    f"git -C {self.path} log -1 --format=%at {commit}",
                ).stdout
            )
        except subprocess.CalledProcessError as e:
            raise RepositoryException(e)

    def apply(self, patches: list[Path], check: bool = False) -> bool:
        patches = [patch.absolute() for patch in patches]
        git_patches = [
            str(patch) for patch in patches if not str(patch).endswith(".sh")
        ]
        sh_patches = [f"sh {patch}" for patch in patches if str(patch).endswith(".sh")]
        if check:
            git_cmd = f"git -C {self.path} apply --check".split(" ") + git_patches
            sh_patches = [patch_cmd + " --check" for patch_cmd in sh_patches]
        else:
            git_cmd = f"git -C {self.path} apply".split(" ") + git_patches

        returncode = 0
        for patch_cmd in [patch_cmd.split(" ") for patch_cmd in sh_patches]:
            returncode += subprocess.run(
                patch_cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
            ).returncode

        if len(git_patches) > 0:
            returncode += subprocess.run(
                git_cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
            ).returncode

        return returncode == 0

    def next_bisection_commit(
        self, good: Revision | Commit, bad: Revision | Commit
    ) -> Commit:
        request_str = (
            f"git -C {self.path} rev-list --bisect --first-parent {bad} ^{good}"
        )
        try:
            return Commit(run_cmd(request_str).stdout)
        except subprocess.CalledProcessError as e:
            raise RepositoryException(e)

    def pull(self) -> None:
        """Pulls from the main branch of the repository.
        It will switch the repository to the main branch.
        It will also invalidate the caches of `rev_to_commit`
        , `get_best_common_ancestor` and `rev_to_tag`.

        Args:
            self:

        Returns:
            None:
        """
        self.rev_to_commit.cache_clear()
        self.get_best_common_ancestor.cache_clear()
        self.rev_to_tag.cache_clear()
        # Just in case...
        cmd0 = f"git -C {self.path} switch {self.main_branch}"
        cmd1 = f"git -C {self.path} pull"
        try:
            run_cmd(cmd0)
            run_cmd(cmd1)
        except subprocess.CalledProcessError as e:
            raise RepositoryException(e)

    @cache
    def rev_to_tag(self, rev: Revision) -> Commit | None:
        request_str = f"git -C {self.path} describe --exact-match {rev}"
        output = subprocess.run(
            request_str.split(),
            capture_output=True,
        )
        stdout = output.stdout.decode("utf-8").strip()
        stderr = output.stderr.decode("utf-8").strip()
        if stderr.startswith("fatal:"):
            return None
        return Commit(stdout)

    @cache
    def parent(self, rev: Revision) -> Commit:
        request_str = f"git -C {self.path} rev-parse {rev}^@"
        try:
            res = run_cmd(request_str).stdout
        except subprocess.SubprocessError as e:
            raise RepositoryException(e)

        assert len(res.split("\n")) == 1
        return Commit(res)

    def prune_worktree(self) -> None:
        prune_str = f"git -C {self.path} worktree prune"
        try:
            run_cmd(prune_str)
        except subprocess.SubprocessError as e:
            raise RepositoryException(e)

    def tags(self) -> list[Revision]:
        print_cmd = f"git -C {self.path} tag -l"
        try:
            res = run_cmd(print_cmd).stdout
        except subprocess.SubprocessError as e:
            raise RepositoryException(e)
        return [Revision(rev) for rev in res.splitlines()]

    def add_worktree(
        self,
        target_path: Path,
        branch: Revision,
        force: bool = True,
        no_checkout: bool = True,
    ) -> None:
        # XXX: this requires manual cleanup
        # TODO: make this a context manager

        cmd = f"git -C {self.path} worktree add {target_path} {branch}"
        if force:
            cmd += " --force"
        if no_checkout:
            cmd += " --no-checkout"

        run_cmd(cmd)

    def remove_worktree(
        self,
        worktree_path: Path,
        force: bool = True,
    ) -> None:
        cmd = f"git -C {self.path} worktree remove {worktree_path}"
        if force:
            cmd += " --force"

        run_cmd(cmd)


def get_llvm_repo(path_to_repo: Path | None = None) -> Repo:
    if path_to_repo:
        return Repo(path_to_repo, Revision("main"))
    return Repo(DEFAULT_REPOS_DIR / "llvm-project", Revision("main"))


def get_gcc_repo(path_to_repo: Path | None = None) -> Repo:
    if path_to_repo:
        return Repo(path_to_repo, Revision("master"))
    return Repo(DEFAULT_REPOS_DIR / "gcc", Revision("master"))


def get_gcc_releases(repo: Repo) -> list[Revision]:
    releases = []
    for r in repo.tags():
        if not r.startswith("releases/gcc-"):
            continue
        # We filter out older releases that we can't build
        should_skip = False
        for v in ("2", "3", "4", "5", "6"):
            if r.startswith(f"releases/gcc-{v}."):
                should_skip = True
                break
        if should_skip:
            continue
        releases.append(r)

    return sorted(
        releases, reverse=True, key=lambda x: int(x.split("-")[-1].replace(".", ""))
    )


def get_llvm_releases(repo: Repo) -> list[Revision]:
    releases = []
    for r in repo.tags():
        if not r.startswith("llvmorg-"):
            continue
        if "-rc" in r or "init" in r:
            continue
        # We filter out older releases that we can't build
        should_skip = False
        for v in ("1", "2", "3", "4"):
            if r.startswith(f"llvmorg-{v}."):
                should_skip = True
                break
        if should_skip:
            continue
        releases.append(r)

    return sorted(
        releases, reverse=True, key=lambda x: int(x.split("-")[-1].replace(".", ""))
    )


def get_releases(project: CompilerProject, repo: Repo) -> list[Revision]:
    match project:
        case CompilerProject.GCC:
            return get_gcc_releases(repo)
        case CompilerProject.LLVM:
            return get_llvm_releases(repo)



from pathlib import Path

import pytest

from diopter.compiler import (
    CompilationSetting,
    CompilerExe,
    CompilerProject,
    Language,
    OptLevel,
    Source,
    SourceFile,
    SourceProgram,
)


def test_setting_serialization() -> None:
    setting0 = CompilationSetting(
        compiler=CompilerExe(CompilerProject.GCC, Path("gcc"), "test"),
        opt_level=OptLevel.O2,
        include_paths=("a", "b"),
        system_include_paths=("sa", "sb"),
        macro_definitions=("M1",),
    )
    setting1 = CompilationSetting(
        compiler=CompilerExe(CompilerProject.LLVM, Path("llvm"), "test"),
        opt_level=OptLevel.O0,
        system_include_paths=("sb",),
    )
    assert setting0 == CompilationSetting.from_json_dict(setting0.to_json_dict())
    assert setting1 == CompilationSetting.from_json_dict(setting1.to_json_dict())


def test_sourceprogram_serialization() -> None:
    p0 = SourceProgram(
        language=Language.C,
        defined_macros=("M1", "M2"),
        include_paths=("a", "b"),
        system_include_paths=("sa", "sb"),
        flags=("-fPIC", "-fno-omit-frame-pointer"),
        code="bla bla",
    )

    assert p0 == SourceProgram.from_json_dict(p0.to_json_dict())
    assert p0 == Source.from_json_dict(p0.to_json_dict())

    p1 = SourceProgram(
        language=Language.CPP,
        defined_macros=("M1",),
        system_include_paths=("sa",),
        code="bla bla blac",
    )

    assert p1 == SourceProgram.from_json_dict(p1.to_json_dict())
    assert p1 == Source.from_json_dict(p1.to_json_dict())

    with pytest.raises(AssertionError):
        SourceFile.from_json_dict(p1.to_json_dict())


def test_sourcefile_serialization() -> None:
    p0 = SourceFile(
        language=Language.C,
        filename=Path("/bla/bla"),
    )

    assert p0 == SourceFile.from_json_dict(p0.to_json_dict())
    assert p0 == Source.from_json_dict(p0.to_json_dict())

    p1 = SourceFile(
        language=Language.CPP,
        defined_macros=("M1",),
        system_include_paths=("sa",),
        filename=Path("bla/bla/blac"),
    )

    assert p1 == SourceFile.from_json_dict(p1.to_json_dict())
    assert p1 == Source.from_json_dict(p1.to_json_dict())

    with pytest.raises(AssertionError):
        SourceProgram.from_json_dict(p1.to_json_dict())



from pathlib import Path

from diopter.compiler import (
    CompilationSetting,
    CompilerExe,
    CompilerProject,
    ObjectCompilationOutput,
    OptLevel,
)
from diopter.generator import CSmithGenerator
from diopter.sanitizer import Sanitizer


def test_preprocessor_make_compiler_agnostic() -> None:
    gcc = CompilerExe(CompilerProject.GCC, Path("gcc"), "")
    clang = CompilerExe(CompilerProject.GCC, Path("clang"), "")

    san = Sanitizer(debug=True)
    gen = CSmithGenerator(san)
    program = gen.generate_program()

    optlevels = [OptLevel.O0, OptLevel.O1, OptLevel.O2, OptLevel.O3, OptLevel.Os]

    for optlevel_preprocess in optlevels:
        for optlevel_test in optlevels:
            gccOp = CompilationSetting(
                compiler=gcc, opt_level=optlevel_preprocess, flags=("-march=native",)
            )
            clangOp = CompilationSetting(
                compiler=clang, opt_level=optlevel_preprocess, flags=("-march=native",)
            )
            gccOt = CompilationSetting(
                compiler=gcc, opt_level=optlevel_test, flags=("-march=native",)
            )
            clangOt = CompilationSetting(
                compiler=clang, opt_level=optlevel_test, flags=("-march=native",)
            )

            pp_with_gcc = gccOp.preprocess_program(program, make_compiler_agnostic=True)
            gccOt.compile_program(
                pp_with_gcc, ObjectCompilationOutput(Path("/dev/null"))
            )
            clangOt.compile_program(
                pp_with_gcc, ObjectCompilationOutput(Path("/dev/null"))
            )
            san.sanitize(pp_with_gcc)

            pp_with_clang = clangOp.preprocess_program(
                program, make_compiler_agnostic=True
            )
            gccOt.compile_program(
                pp_with_clang, ObjectCompilationOutput(Path("/dev/null"))
            )
            clangOt.compile_program(
                pp_with_clang, ObjectCompilationOutput(Path("/dev/null"))
            )
            san.sanitize(pp_with_clang)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from subprocess import CalledProcessError
from tempfile import TemporaryDirectory

from diopter.repository import Commit, Repo, Revision
from diopter.utils import run_cmd


def rev_parse(worktree_dir: Path, rev: str) -> Commit:
    return Commit(run_cmd(f"git -C {worktree_dir} rev-parse {rev}").stdout)


def parse_bisect_head(worktree_dir: Path) -> Commit:
    return Commit(rev_parse(worktree_dir, "BISECT_HEAD"))


def get_current_bisection_commit(working_dir: Path, no_checkout: bool) -> Commit:
    if no_checkout:
        return parse_bisect_head(working_dir)
    else:
        return rev_parse(working_dir, "HEAD")


def bisect_start(
    worktree_dir: Path,
    bad: Revision | Commit,
    good: Revision | Commit,
    no_checkout: bool = True,
) -> Commit:
    # TODO: make this a context manager?
    cmd = (
        f"git -C {worktree_dir} bisect start --first-parent"
        + (" --no-checkout" if no_checkout else "")
        + f" {bad} {good}"
    )
    output = run_cmd(cmd)
    print(output.stdout)
    return get_current_bisection_commit(worktree_dir, no_checkout)


def bisect_skip(worktree_dir: Path, commit: Commit) -> None:
    cmd = f"git -C {worktree_dir} bisect skip"
    if commit is not None:
        cmd += f" {commit}"
    print(run_cmd(cmd).stdout)


def bisect_good(worktree_dir: Path, commit: Commit) -> None:
    cmd = f"git -C {worktree_dir} bisect good"
    if commit is not None:
        cmd += f" {commit}"
    print(run_cmd(cmd).stdout)


def bisect_bad(worktree_dir: Path, commit: Commit) -> None:
    cmd = f"git -C {worktree_dir} bisect bad"
    if commit is not None:
        cmd += f" {commit}"
    print(run_cmd(cmd).stdout)


def bisect_log(worktree_dir: Path) -> str:
    return run_cmd(
        f"git -C {worktree_dir} bisect log",
    ).stdout


def latest_good_commit(worktree_dir: Path) -> Commit:
    good_prefix = "# good: ["
    good_commits = []
    for line in bisect_log(worktree_dir).splitlines():
        if line.startswith(good_prefix):
            good_commits.append(line[len(good_prefix) :].split("]", 1)[0])

    assert len(good_commits) > 0
    return Commit(good_commits[-1])


def currently_bisecting(worktree_dir: Path) -> bool:
    return (
        not bisect_log(worktree_dir).splitlines()[-1].startswith("# first bad commit")
    )


def successful_bisection(worktree_dir: Path) -> bool:
    return bisect_log(worktree_dir).splitlines()[-1].startswith("# first bad commit")


@dataclass(frozen=True)
class BisectTestResult:
    commit: Commit
    is_good: bool | None


class BisectionCallback(ABC):
    def shift_tested_commit(
        self, commit: Commit, good_commit: Commit, bad_commit: Commit
    ) -> Commit:
        """Shifts the commit to be tested.
        This can be used to steer the bisection
        """
        return commit

    @abstractmethod
    def check_impl(self, commit: Commit, repo_dir: Path) -> bool | None:
        """Checks if the commit is good or bad or broken.

        Subclasses should implement this method.
        """
        pass

    def check(
        self, commit: Commit, good_commit: Commit, bad_commit: Commit, repo_dir: Path
    ) -> BisectTestResult:
        """Checks if the commit is good or bad or broken."""
        commit = self.shift_tested_commit(commit, good_commit, bad_commit)
        result = self.check_impl(commit, repo_dir)
        if result is None:
            return BisectTestResult(commit, None)
        return BisectTestResult(commit, result is False)


def bisect(
    repo: Repo,
    callback: BisectionCallback,
    no_checkout: bool = True,
    *,
    good: Revision | Commit,
    bad: Revision | Commit,
) -> Commit | None:
    # TODO: add support for specifying which paths in the repo to look at
    with TemporaryDirectory() as tempdir:
        worktree_dir = Path(tempdir)
        # The context manager should be the worktree
        repo.add_worktree(worktree_dir, repo.current_branch(), no_checkout=no_checkout)
        bisect_start(worktree_dir, bad, good, no_checkout=no_checkout)
        while currently_bisecting(worktree_dir):
            commit = get_current_bisection_commit(worktree_dir, no_checkout)
            test_result = callback.check(
                commit,
                latest_good_commit(worktree_dir),
                rev_parse(worktree_dir, "bisect/bad"),
                worktree_dir,
            )
            print("Testing:", commit)
            if test_result.commit != commit:
                print(f"Bisection commit shifted: {commit} -> {test_result.commit}")
            try:
                if test_result.is_good is None:
                    print(f"Skipping {commit}")
                    bisect_skip(worktree_dir, test_result.commit)
                    continue
                if test_result.is_good:
                    print(f"Good {commit}")
                    bisect_good(worktree_dir, test_result.commit)
                else:
                    print(f"Bad {commit}")
                    bisect_bad(worktree_dir, test_result.commit)

            except CalledProcessError as e:
                print(e.stdout.decode("utf-8"))
                return None
        if not successful_bisection(worktree_dir):
            return None
        commit = rev_parse(worktree_dir, "bisect/bad")
        repo.remove_worktree(worktree_dir)

    return commit
